/*
 * Copyright 2011 Nate Koenig & Andrew Howard
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/
/* Desc: The base class for all physics engines
 * Author: Nate Koenig
 */

#include "msgs/msgs.h"
#include "common/Exception.hh"
#include "common/Console.hh"
#include "common/Events.hh"

#include "transport/Transport.hh"
#include "transport/Node.hh"

#include "physics/World.hh"
#include "physics/PhysicsEngine.hh"

#include "sdf/sdf_parser.h"

using namespace gazebo;
using namespace physics;

////////////////////////////////////////////////////////////////////////////////
// Constructor
PhysicsEngine::PhysicsEngine(WorldPtr _world)
  : world(_world)
{
  this->sdf.reset(new sdf::Element);
  sdf::initFile( "/sdf/physics.sdf", this->sdf );

  this->node = transport::NodePtr(new transport::Node());
  this->node->Init(this->world->GetName());
  this->physicsSub = this->node->Subscribe("~/physics",
      &PhysicsEngine::OnPhysicsMsg, this);

  this->responsePub = this->node->Advertise<msgs::Response>("~/response");
  this->requestSub = this->node->Subscribe("~/request", 
                                           &PhysicsEngine::OnRequest, this);

  this->rayMutex = new boost::recursive_mutex();
}

void PhysicsEngine::Fini()
{
  this->world.reset();
  this->node->Fini();
}

////////////////////////////////////////////////////////////////////////////////
// Destructor
PhysicsEngine::~PhysicsEngine()
{
  this->sdf->Reset();
  this->sdf.reset();
  delete this->rayMutex;
  this->rayMutex = NULL;
  this->responsePub.reset();
  this->requestSub.reset();
  this->node.reset();
}

////////////////////////////////////////////////////////////////////////////////
/// Return the gavity vector
math::Vector3 PhysicsEngine::GetGravity() const
{
  return this->sdf->GetOrCreateElement("gravity")->GetValueVector3("xyz");
}
